"""A wrapper generator for IPyWidgets, allows interop with MavenWorks."""

from .KernelPart import KernelPart, register_part
from ..serialization import traitlet_to_type
import ipywidgets as widgets
import traitlets
from IPython.core.display import display


def gen_wrapper(widget_cls, trait_names):
    """Generate a widget wrapper and register it immediately.

    :param widget_cls: The Widget to wrap into a Part
    :param trait_names: A list of widget properties that should become part
    options
    """
    @register_part(widget_cls.__name__ + "Part")  # pylint: disable=W0612
    class WrapperPart(KernelPart):
        @classmethod
        def get_metadata(cls):
            metadata = super().get_metadata()
            traits = widget_cls.class_traits()
            for opt in trait_names:
                trait = traits[opt]
                default = trait.default_value
                if default is traitlets.Undefined:
                    default = None
                type_annotation = traitlet_to_type(trait)
                metadata.add_option(opt, default, type_annotation)
            return metadata

        def initialize(self):
            self.widget = widget_cls()
            self.bag = None
            self.widget.observe(self._on_widget_updated, names=trait_names)

        def _on_widget_updated(self, change):
            if self.bag is None:
                return

            self.bag[change["name"]] = change["new"]

        def render(self, opts):
            self.bag = None
            for opt in opts:
                setattr(self.widget, opt, opts[opt])
            self.bag = opts
            display(self.widget)


# An example
gen_wrapper(widgets.IntSlider, [
    "value",
    "min",
    "max",
    "step",
    "description",
    "readout",
    "readout_format"
])



"""Handler for the Dashbook viewer."""

from notebook.base.handlers import IPythonHandler
import os


class ViewerApp(IPythonHandler):
    """Renders the viewer itself, with page config overrides."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get(self, path=""):
        page_config = self.settings.get("page_config_data", {
            "wsUrl": self.ws_url
        })
        page_config.update({"baseUrl": self.base_url})
        return self.render(
            "viewer.html",
            baseUrl=self.base_url,
            page_config=page_config,
            # easter egg to use the cooler loading icon in place of the spinny
            not_boring=(os.getenv("COOLER_LOADING_ICON", None) is not None)
        )

    def get_template_path(self):
        return os.path.abspath(
            os.path.join(os.path.dirname(__file__), "./static/")
        )


class ViewerIndex(IPythonHandler):
    """Sends an index for conveniently navigating among the files."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get(self, path=""):
        folder = self.contents_manager.get(path)
        if folder is None or folder['type'] != "directory":
            self.set_status(404)
            self.finish()
            return
        objs = folder['content']
        folders = list(filter(lambda i: i['type'] == "directory", objs))
        notebooks = list(filter(lambda i: i['type'] == "notebook", objs))
        self.render(
            "index.html",
            baseUrl=self.base_url,
            folders=folders,
            notebooks=notebooks
        )

    def get_template_path(self):
        return os.path.abspath(
            os.path.join(os.path.dirname(__file__), "./static/")
        )



from .part import Part
from .helpers import kwargs_to_properties


class Container:
    def __init__(self, container_type, *children, **properties):
        self.__type = container_type
        self.__children = children or []
        self.__properties = properties or {}
        self._name = "Container"

    def add(self, *children):
        self.__children += children

    @property
    def properties(self):
        return self.__properties

    def __repr__(self):
        repr_str = self._name + ":"
        idx = 0
        for child in self.__children:
            idx += 1
            child_repr = repr(child)
            repr_str += "\n+ Child[" + str(idx) + "]:"
            repr_str += "\n|".join(
                map(lambda i: "\t" + i, child_repr.split("\n"))
            )
        return repr_str

    def _collect_parts(self):
        parts = []
        for child in self.__children:
            if isinstance(child, Part):
                parts.append(child)
            elif isinstance(child, Container):
                parts += child._collect_parts()
        return parts

    def _as_json(self):
        child_attached_props = []
        children = []
        attached_props, layout_props = kwargs_to_properties(self.__properties)
        for child in self.__children:
            representation, child_props = child._as_json()
            children.append(representation)
            child_attached_props.append(child_props)

        return {
            "properties": layout_props,
            "attachedProperties": child_attached_props,
            "typeName": self.__type,
            "children": children
        }, attached_props


class StackPanel(Container):
    def __init__(self, *children, **properties):
        super().__init__(0, *children, **properties)


class TabPanel(Container):
    def __init__(self, *children, **properties):
        super().__init__(2, *children, **properties)


class CanvasPanel(Container):
    def __init__(self, *children, **properties):
        super().__init__(3, *children, **properties)


class GridPanel(Container):
    def __init__(self, *children, **properties):
        super().__init__(4, *children, **properties)



from ..parts import KernelPart, OptionsBag
from IPython.utils.capture import capture_output
from IPython.core.ultratb import VerboseTB
from ipykernel.comm import Comm, CommManager
from IPython.core.getipython import get_ipython
from IPython.core.formatters import format_display_data
from ..serialization import serialize, guess_type, deserialize
import sys


class KernelPartManager:
    instance = None

    @classmethod
    def Create(cls):
        if cls.instance is not None:
            return cls.instance
        return KernelPartManager()

    def __init__(self):
        if self.instance is not None:
            raise RuntimeError(
                "Only one instance of KernelPartManager is allowed"
            )
        self.parts = {}
        self.options_bags = {}
        self.error_formatter = VerboseTB()
        # Connect to a comm

    def create_part(self, type_name, uuid):
        part = KernelPart.Create(type_name)
        part.uuid = uuid
        self.parts[uuid] = part
        self.options_bags[uuid] = OptionsBag(part.get_metadata())

    def destroy_part(self, uuid):
        self.parts[uuid].dispose()
        del self.parts[uuid]
        del self.options_bags[uuid]

    def initialize_part(self, uuid):
        self.parts[uuid].initialize()

    def render_part(self, uuid, options):
        bag: OptionsBag = self.options_bags[uuid]
        bag.is_stale = True
        for opt in options.keys():
            bag[opt] = options[opt]
        bag.set_fresh()
        with capture_output() as capture:
            ret = self.parts[uuid].render(bag)
        if ret is not None:
            return ret
        elif len(capture.outputs) > 0:
            return capture.outputs[0]
        else:
            return capture.stdout + capture.stderr

    def dispatch_msg(self, msg, comm: Comm):
        data = msg['content']['data']
        msg_type = data['msg_type']
        uuid = data['uuid']
        payload = data['payload']
        if msg_type == "create":
            self.create_part(payload, uuid)
            bag: OptionsBag = self.options_bags[uuid]
            bag.OnStale.subscribe(lambda args: comm.send({
                "msg_type": "stale",
                "uuid": uuid,
                "payload": {
                    "name": args[0],
                    "value": serialize(args[1], guess_type(args[1]))
                }
            }))
        if msg_type == "initialize":
            error = None
            try:
                self.initialize_part(uuid)
            except:  # noqa: E722
                exc_info = sys.exc_info()
                error = self.error_formatter.text(*exc_info)
            comm.send({
                "msg_type": "initialize_done",
                "uuid": uuid,
                "payload": None,
                "error": error
            })
        if msg_type == "render":
            options = {
                name: deserialize(value) for name, value in payload.items()
            }
            error = None
            value = None
            try:
                display_data, display_metadata = format_display_data(
                    self.render_part(uuid, options)
                )
                value = {
                    "data": display_data,
                    "metadata": display_metadata
                }
            except:  # noqa: E722
                exc_info = sys.exc_info()
                error = self.error_formatter.text(*exc_info)
            comm.send({
                "msg_type": "render_done",
                "uuid": uuid,
                "payload": value,
                "error": error
            })
        if msg_type == "dispose":
            return self.destroy_part(uuid)


manager = KernelPartManager.Create()
ip = get_ipython()


def _register_new_client(comm, msg):
    comm.on_msg(lambda msg: manager.dispatch_msg(msg, comm))
    # TODO: Track what parts were created by a comm and make sure all those
    # parts are destroyed when the channel dies


if ip is not None:
    comm_manager: CommManager = ip.kernel.comm_manager
    comm_manager.register_target("kernel_proxy_part", _register_new_client)



"""Python Cell Dashboard API.

This module provides an API for writing MavenWorks dashboards using Python. It
works in JupyterLab, using Jupyter-native machinery. Because it doesn't use the
same tooling as 'normal' dashboards, some editors (like the Visual Editor)
won't work with these types of dashboards.
"""

from . import bind as Bind
from .dashboard import Dashboard
from .containers import StackPanel, TabPanel, GridPanel, CanvasPanel
from .part import Part

__all__ = [
    "Bind",
    "Dashboard",
    "StackPanel",
    "TabPanel",
    "GridPanel",
    "CanvasPanel",
    "Part"
]



"""Internal services for MavenWorks in Python."""
# can't do `import .pkg-name` in Python
from .expression_evaluator import *  # noqa F401 F403
from .KernelPartManager import *  # noqa F401 F403
from .MetadataComm import *  # noqa F401 F403
from IPython.core.display import display

display({
    "text/plain": ""
}, transient={
    "mavenomics_state_ok": True
}, raw=True)



"""Kernel Parts."""

from uuid import uuid4
from .PartHelpers import PartMetadata
from typing import Union, Callable, ItemsView

registry = {}
cb = None


def _set_new_part_hook(callback: Union[Callable, None]):
    global cb
    cb = callback


def _get_all_parts() -> ItemsView:
    return registry.items()


def register_part(name=None):
    """Register a KernelPart with MavenWorks.

    :param name: A name for this part, defaults to class name
    :type name: str, optional

    :Example:

    >>> @register_part()
    ... class MyCoolPart(KernelPart):
    ...    pass
    """
    def register(f):
        # fun fact: this is only sort of a closure because of how it's called
        # so we need the nonlocal keyword to tell Python that this is a closure
        nonlocal name
        if name is None:
            name = f.__name__
        registry[name] = f
        if cb is not None:
            cb(name, f)
        return f
    return register


class KernelPart():
    """Abstract base class for KernelParts.

    :Example:

    >>> import time
    >>> @register_part()
    ... class MyPart(KernelPart):
    ...     @classmethod()
    ...     def get_metadata(cls):
    ...         metadata = super().get_metadata()
    ...         metadata.add_option("NumberToSquare", "Number", 42)
    ...         return metadata
    ...     def initialize(self):
    ...         display("Hello, world!")
    ...     def render(self, opts):
    ...         number_to_square = opts["NumberToSquare"]
    ...         # The user will see a 'Rendering' overlay while ``render``
    ...         # runs, unless [[showOverlays]] is false.
    ...         time.sleep(1)
    ...         display(number_to_square ** 2)
    ...

    """

    @staticmethod
    def Create(name):
        """Do not use. Internal method for the KernelPartManager."""
        part = registry[name]()
        return part

    @classmethod
    def get_metadata(cls):
        """Describe this part's metadata.

        ..note::
            KernelParts should override this method to add options.

        :Example:

        >>> class MyPart(KernelPart):
        ...     @classmethod
        ...     def get_metadata(cls):
        ...         metadata = super().get_metadata()
        ...         metadata.add_option("MyFooOption", "Number", 42)
        ...         return metadata
        """
        return PartMetadata()

    def __init__(self):
        """Part constructor."""
        self.uuid = str(uuid4())
        self.is_disposed = False

    def initialize(self):
        """Initialize this part, performing any necessary setup.

        .. note::
            ``initialize`` is called by the framework after everything on the
            client is setup. Use this method for long-running tasks that only
            need to run once, such as fetching dependencies or setting up
            initial view state.
        """
        pass

    def render(self, options):
        """Render this part.

        :param options: An OptionsBag holding the values of this parts options
        :type options: OptionsBag

        .. note::
            ``render`` is called by the framework after an option has changed.
            This means it can be called frequently, and therefore should be
            as fast as possible.

        .. note::
            Anything ``display()``'d or ``return``'d from this method will be
            displayed on the client. Part writers do not need to worry about
            how the output will be displayed.
            If there is no output, then ``stdout`` and ``stderr`` will be
            displayed instead.
        """
        pass

    def dispose(self):
        """Clean up any resources this part held.

        .. note::
            This method is recommended over ``__del__`` since the framework
            calls it explicitly upon part deconstruction, and will not
            introduce the harmful GC loops that ``__del__`` can.
        """
        if self.is_disposed:
            return
        self.is_disposed = True



"""Provides the Part wrapper for the Textual Declarative API."""

import uuid
from ..serialization import serialize, guess_type
from .bind import Binding
from .helpers import kwargs_to_properties


def serialize_opt(opt):
    """Private helper for serializing a part binding."""
    if isinstance(opt, Binding):
        return opt._as_json()
    return serialize(opt, guess_type(opt))


class Part:
    MIMETYPE = "application/vnd.maven.part+json"

    def __init__(self, name, options, **props):
        self.__name = name
        self.__options = options
        self.__props = props
        self.__uuid = str(uuid.uuid4())
        self.__layout_uuid = str(uuid.uuid4())

    def _as_part(self):
        return {
            "text/plain": self.__name + " Part Model",
            Part.MIMETYPE: {
                "name": self.__name,
                "id": self.__uuid,
                "options": {
                    opt: serialize_opt(value)
                    for opt, value in self.__options.items()
                }
            }
        }, self.__uuid

    def _as_json(self):
        attached_props, layout_props = kwargs_to_properties(self.__props)
        return {
            "guid": self.__uuid,
            "id": self.__layout_uuid,
            "typeName": 1,
            "properties": layout_props
        }, attached_props



"""MavenWorks Dashboarding Framework."""

__version__ = "0.1.0"

from .parts import gen_wrapper, KernelPart, name_display_handle,\
    register_part, wrap, Option, OptionsBag
from .serialization import guess_type, serialize, deserialize
from .dashboard import Bind, Dashboard, StackPanel, TabPanel, GridPanel, \
    CanvasPanel, Part
from .services import *  # noqa F401 F403

__all__ = [
    "name_display_handle",
    "register_part",
    "KernelPart",
    "gen_wrapper",
    "wrap",
    "guess_type",
    "serialize",
    "deserialize",
    "Option",
    "OptionsBag",
    "gen_wrapper",
    "Bind",
    "Dashboard",
    "StackPanel",
    "TabPanel",
    "GridPanel",
    "CanvasPanel",
    "Part",
]



"""Helper for loading config objects."""

import os
import json
import warnings
from ..settings import get_setting

parts_dir = get_setting("global_parts_folder")


class ConfigObjectTracker:
    """Class to sniff all the config objects in a directory.

    For the first cut, there's no file watching or update msgs. We could add
    it, but for now YAGNI applies as there's no need for any editors to be
    aware of config staleness (as they just use file i/o directly).

    The use case for update msgs would be for updating a downstream cache,
    which is a likely future optimization (especially if we needed to use
    this in large-scale deployments- the current approach of checking on every
    query works is just fine for small-scale and individual deployments).
    """

    @staticmethod
    def get_objects_in_tree(root_dir, cwd):
        """Search for config objects, given a root directory and a scoped path.

        Scoping allows config objects to not 'poison' unrelated work in other
        directories. The scoping is rather primitive, and based on the folder
        tree. If anything appears in the global parts directory, it will appear
        to all dashboards. If a config object appears in the scoped directory,
        it will appear to all dashboards under that scope.

        If a config object in local scope has the same name as a config object
        in global scope, the local object will be ordered after the global
        object in the response. Clients should therefore "shadow" the global
        object, replacing it with the local object.
        """
        root_dir = os.path.abspath(root_dir)
        # strip any leading slash
        cwd = cwd[1:] if cwd.startswith("/") else cwd
        work_dir = os.path.abspath(cwd)
        if not work_dir.startswith(root_dir):
            msg = "Working directory must be under Jupyter root! Got: "
            msg += work_dir
            raise Exception(msg)
        if (work_dir != root_dir):
            yield from ConfigObjectTracker.get_objects_in_subtree(
                root_dir,
                parts_dir
            )
        yield from ConfigObjectTracker.get_objects_in_subtree(root_dir, cwd)

    @staticmethod
    def get_objects_in_subtree(root_dir, subtree_dir):
        """Get all config objects in a given subtree, correcting for path."""
        actual_path = os.path.join(root_dir, subtree_dir)
        with os.scandir(actual_path) as child_iter:
            for entry in child_iter:
                if entry.name.startswith("."):  # Ignore hidden files+dirs
                    continue
                success, obj = ConfigObjectTracker.__try_read_file(
                    entry,
                    actual_path,
                    root_dir
                )
                if not success:
                    if entry.is_dir():
                        yield from ConfigObjectTracker.get_objects_in_subtree(
                            root_dir,
                            entry.path
                        )
                    continue
                yield obj

    @staticmethod
    def __try_read_file(dir_entry: os.DirEntry, scan_dir: str, base_path: str):
        """Attempt to read a file. Internal.

        If the file given by ``dir_entry`` is a Maven config object, return
        a tuple of (True, JSON dict) representing that object. If the object
        could be read but wasn't a part, return (False, None). If the object
        couldn't be read, raise an exception.

        :param dir_entry: A DirEntry object, as returned by os.scandir
        :param scan_dir: The directory on which os.scandir was called. May be
                         an empty string.
        :param base_path: The base path of the config, where "/" is the root of
        the config.
        """
        name, ext = os.path.splitext(dir_entry.path)
        if dir_entry.is_dir() or ext != ".part":
            return (False, None)
        file_path = os.path.join(base_path, scan_dir, dir_entry.path)
        try:
            with open(file_path, 'r') as f:
                data = json.load(f)
                path = ConfigObjectTracker.__norm_path(file_path, base_path)
                # return a corrected dict
                obj = {
                    # todo: Should we generate a uuid1 from the file's inode?
                    # Technically IDs should be unique to a config and
                    # immutable across a given object's lifetime, however
                    # inodes are only unique to a block device and may change
                    # with renames/moves. This means a config with a symlink to
                    # a network drive or another disk might have duplicate
                    # inodes, and with uuid1s, they will have duplicate IDs.
                    "id": data["id"],
                    "typeName": data["typeName"],
                    # Because this is a directory store, we can safely ignore
                    # whatever the file says is true, because the filesystem
                    # is more accurate in these regards
                    "lastModified": os.path.getmtime(file_path),
                    "name": os.path.basename(name),
                    "path": path,
                }
                if "data" in data:
                    # this is a FileConfigObject, attach the data
                    obj["data"] = data["data"]
                if "functionType" in data and "arguments" in data:
                    # this is a UDF or UDP, attach those args
                    obj["functionType"] = data["functionType"]
                    obj["arguments"] = data["arguments"]
                return (True, obj)
        except (json.JSONDecodeError, KeyError):
            # Not a part, or a malformed part
            return (False, None)
        except UnicodeDecodeError as e:
            # Corrupted file, could not read
            warnings.warn("Failed to read file" + file_path)
            warnings.warn(e)
            return (False, None)
        # Other exception classes are raised to the parent, like decoding
        # errors or permission errors. These indicate that the file can't
        # be read, not that the file isn't a part.

    @staticmethod
    def __norm_path(old_path, config_base_path):
        r"""Normalize the path to the canonical form of a Maven config path.

        This means that Windows-style slashes ('\') are replaced with UNIX-
        style slashes, trailing slashes are removed, and a leading slash is
        added. The path given is relative to ``config_base_bath``, which is the
        root directory of the config.
        """
        fixed_path = os.path.relpath(old_path, config_base_path)
        fixed_path = fixed_path.replace("\\", "/")
        if not fixed_path.startswith("/"):
            fixed_path = "/" + fixed_path
        if fixed_path.endswith("/") and len(fixed_path) > 1:
            fixed_path = fixed_path[1:]
        return fixed_path



"""Setup file for MavenWorks python package."""

from setuptools import setup
from setuptools.command.install import install
from subprocess import check_call
import os


# Many thanks to https://stackoverflow.com/a/36902139
class PostInstallCommand(install):
    """Post-installation for installation mode."""
    def run(self):
        install.run(self)
        check_call("jupyter serverextension enable --py mavenworks.server".split())

with open("README.md") as f:
    desc = f.read()

setup(
    name="mavenworks",
    version="0.0.4",
    description="Dashboarding for JupyterLab",
    long_description=desc,
    long_description_content_type="text/markdown",
    author="Mavenomics, Inc",
    license="GPL-3.0 (C) 2019 Mavenomics, Inc.",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Framework :: IPython",
    ],
    cmdclass={
        'install': PostInstallCommand,
    },
    packages=[
        "mavenworks",
        "mavenworks.dashboard",
        "mavenworks.parts",
        "mavenworks.services",
        "mavenworks.server",
    ],
    install_requires=[
        "requests",
        "rx>=1.0,<3",
        "jupyterlab>1.0",
        "pandas",
        "IPython",
        "ipywidgets",
        "matplotlib",
    ],
    extra_requires={
        "pyarrow-table-interop": [
            "pyarrow"
        ]
    },
    python_requires="~=3.6",
    include_package_data=True
)



"""
KernelExpressionEvaluator
=========================

MavenWorks utility for evaluating global expressions

_Note_: The "right" way to get globals referenced by an expression  is to have
this comm do it. But as things are right now, Python's AST module doesn't work
well with modified grammars. So for now the client handles this with a hacky
regex. Not ideal, but it works for now.

Expressions are kernel code that reference global values and form a reactive
calculation (like Excel formulas). Clients track globals, and when those
globals change, the client will re-evaluate the expression by sending a message
to this comm with the current values of the referenced globals.

The comm accepts the following messages:

 - ``evaluate_expr``: Given an expression and the values of a set of globals,
   evaluate the expression and return the result. If the expression failed to
   complete, trap the error and return it to the front-end

The comm sends the following messages:

 - ``expr_value``: Returns the value of an expression given by a previous
   ``evaluate_expr`` message. If there was an error in evaluation, this comm
   will instead send ``expr_error``.
 - ``expr_error``: If a expression evaluation failed, this will be sent instead
   of ``expr_value`` and will include a serialized form of error that clients
   must present to the user.
"""


import re
import sys
from typing import AnyStr, Dict, Any
from IPython import get_ipython
from IPython.core.ultratb import VerboseTB
from ipykernel.comm import Comm
from ..serialization import serialize, deserialize, guess_type

MESSAGE_TYPES = [
    "evaluate_expr"
]
tb_formatter = VerboseTB()
global_regex = re.compile(r"\@([A-Za-z][A-Za-z0-9_]*)")


def evaluate_expr(
        expr: AnyStr,
        globals_dict: Dict[AnyStr, Any],
        comm: Comm,
        parent: AnyStr):
    ip = get_ipython()
    locals_dict = {}
    locals_dict.update(ip.user_ns)
    locals_dict.update(globals_dict)
    expr = global_regex.sub(r"\1", expr)
    try:
        value = eval(expr, ip.user_global_ns, locals_dict)
    except:  # noqa: E722
        exc_info = sys.exc_info()
        exc = tb_formatter.text(*exc_info)
        comm.send({
            "msg_type": "expr_error",
            "payload": serialize(exc, "String"),
            "parent": parent
        })
        return
    comm.send({
        "msg_type": "expr_value",
        "payload": serialize(value, guess_type(value)),
        "parent": parent
    })


def dispatch_message(comm, msg):
    content = msg["content"]["data"]
    msg_type = content["msg_type"]
    if msg_type not in MESSAGE_TYPES:
        raise KeyError("Unrecognized message type " + msg_type)
    globals_dict = {
        g: deserialize(v) for g, v in content["globals"].items()
    }
    evaluate_expr(content.get("expr", ""), globals_dict, comm, content["uuid"])


def register_frontend(comm, _msg):
    comm.on_msg(lambda msg: dispatch_message(comm, msg))


ip = get_ipython()

if ip is not None:
    ip.kernel.comm_manager.register_target(
        "expression_evaluator",
        register_frontend
    )
